<?php

namespace Ems\Services;

use Util\DateUtil;
use Util\Environment;
use Util\Messages;

use Ems\Util\Utils;
use Ems\Repository\{ MarcacaoRepository, ApuracaoRepository, PeriodoRepository };
use Ems\Services\{ OcorrenciaService, ApuracaoService, VinculoService };

use Doctrine\ORM\EntityManager;

class MarcacaoPontoService {
    
    const IDTIPOINTERVALO_TRABALHO          = 'TRABALHO';
    const LABEL_IDTIPOINTERVALO_TRABALHO    = 'Trabalho';
    const IDTIPOINTERVALO_AUSENCIA          = 'AUSENCIA';
    const LABEL_IDTIPOINTERVALO_AUSENCIA    = 'Ausência';
    const TIPO_VINC_CLT                     = 1;
    const VINC_EMPREG_APRENDIZ              = 9;

    public function __construct(
        VinculoService      $vinculoService,
        MarcacaoRepository  $marcacaoRepository,
        ApuracaoRepository  $apuracaoRepository,
        PeriodoRepository   $periodoRepository,
        EntityManager       $entityManager,
        Environment         $environment
    ) {
        $this->vinculoService       = $vinculoService;
        $this->marcacaoRepository   = $marcacaoRepository;
        $this->apuracaoRepository   = $apuracaoRepository;
        $this->periodoRepository    = $periodoRepository;
        $this->entityManager        = $entityManager;
        $this->environment          = $environment;
    }

    public function getMarcacoesPonto($parameters) {

        $params = [
            'P_NRORG'           => $parameters['P_NRORG'],
            'P_NRVINCULOM'      => $parameters['P_NRVINCULOM'],
            'P_DTAPURACAOPONTO' => $parameters['P_DTAPURACAOPONTO']
        ];

        $sql = <<<SQL
            SELECT
              ROWNUM,
              NRMARCACAOPONTO,
              MARC.NRVINCULOM,
              TO_CHAR(DTHRAJUSTADA, 'DD/MM/YYYY HH24:MI:SS') AS DTHRAJUSTADA,
              TO_CHAR(DTHRAJUSTADA, 'DD/MM/YYYY') AS DTMARCACAOPONTO,
              TO_CHAR(DTHRAJUSTADA, 'HH24:MI') AS HORA_MINUTO,
              DTHRAJUSTADA AS DTHRORDER,
              IDMARCDESPREZADA,
              IDDIAOCORRENCIA,
              DSJUSTIFICATIVA,
              COL.NRCOLETOR,
              COL.DSCOLETOR,
              IDREFERINTERVALO
             FROM GPE_MARCACAOPONTO MARC
             JOIN GPE_COLETOR COL
               ON COL.NRCOLETOR = MARC.NRCOLETOR
            WHERE MARC.NRORG = :P_NRORG
              AND MARC.NRCOLETOR != 21
              AND MARC.NRCOLETOR != 25
              AND MARC.NRVINCULOM = :P_NRVINCULOM
              --AND MARC.IDMARCDESPREZADA = 'N'
              AND (((IDDIAOCORRENCIA IS NULL OR IDDIAOCORRENCIA = 'DC')
                     AND TO_DATE(TO_CHAR(DTHRAJUSTADA,'DD/MM/YYYY')) = :P_DTAPURACAOPONTO)
                   OR (IDDIAOCORRENCIA IN ('FA', 'DA') AND TRUNC(DTHRAJUSTADA) = TO_DATE(:P_DTAPURACAOPONTO,'DD/MM/YYYY') + 1)
                   OR (IDDIAOCORRENCIA IN ('FS', 'DS') AND TRUNC(DTHRAJUSTADA) = TO_DATE(:P_DTAPURACAOPONTO,'DD/MM/YYYY') - 1)
                   OR (IDDIAOCORRENCIA = 'FC' AND TRUNC(DTHRAJUSTADA) = TO_DATE(:P_DTAPURACAOPONTO,'DD/MM/YYYY')))
            ORDER BY DTHRAJUSTADA, NRMARCACAOPONTO
SQL;
        $query = $this->entityManager->getConnection()->prepare($sql);

        $query->execute($params);

        $marcacoesPonto = $query->fetchAll();
        $nrvinculom = $parameters['P_NRVINCULOM'];
        $intervalos = $this->factoryIntervaloMarcacoes($marcacoesPonto, $parameters["P_DTAPURACAOPONTO"], $nrvinculom);

        if (isset($intervalos)) {
            return $intervalos;
        } else {
            return null;
        }
    }

    public function factoryIntervaloMarcacoes($marcacoesPonto, $dtapuracaoponto, $nrvinculom) {

        $intervalosHorario = array();

        // gera os intervalos de entrada e calcula o tempo
        $intervalosEntrada = $this->getIntervalosHorario($marcacoesPonto, self::LABEL_IDTIPOINTERVALO_TRABALHO, $dtapuracaoponto, $nrvinculom);

        // remove as marcações de primeira entrada (chegada) e última sáida (embora)
        unset($marcacoesPonto[0]);
        array_pop($marcacoesPonto);

        // gera os intervalos de saída e calcula o tempo
        $intervalosSaida = $this->getIntervalosHorario($marcacoesPonto, self::LABEL_IDTIPOINTERVALO_AUSENCIA, $dtapuracaoponto, $nrvinculom);

        // reordena o array pelo o horário, colocando-o como chave para facilitar a ordenação.
        $pares = [];
        $qtEntradas = count($intervalosEntrada);

        for ( $i = 0; $i < $qtEntradas; $i++) {
            array_push($pares, $intervalosEntrada[$i]);

            if (isset($pares, $intervalosSaida[$i])) {
                array_push($pares, $intervalosSaida[$i]);
            }
        }

        return $pares;
    }

    private function getIntervalosHorario($marcacoesPonto, $idmarcacao, $dtapuracaoponto, $nrvinculom) {

        $paresMarcacoes = array_chunk($marcacoesPonto, 2);

        $intervalos = array();

        

        foreach ($paresMarcacoes as $key => $intervalosMarcacoes) {
            /* Posição 0 do array "$intervalosMarcacoes" são marcações de início
             * Posição 1 do array "$intervalosMarcacoes" são marcações de fim
             */

            // garante que foi retornado um intervalo com duas marcações
            if (count($intervalosMarcacoes) == 2) {
                $intervalos[$key]["NRMARCACAOPONTO"]       = $intervalosMarcacoes[0]["NRMARCACAOPONTO"];
                $intervalos[$key]["NRMARCACAOPONTO_SAIDA"] = $intervalosMarcacoes[1]["NRMARCACAOPONTO"];

                $intervalos[$key]["IDMARCACAO"]      = $idmarcacao;
                $intervalos[$key]["DTMARCACAOPONTO"] = $intervalosMarcacoes[0]["DTMARCACAOPONTO"];

                $intervalos[$key]["DTHRINICIO"] = $intervalosMarcacoes[0]["HORA_MINUTO"];
                $intervalos[$key]["DTHRFIM"]    = $intervalosMarcacoes[1]["HORA_MINUTO"];

                $intervalos[$key]["DIA_INICIO"] = $intervalosMarcacoes[0]["IDDIAOCORRENCIA"];
                $intervalos[$key]["DIA_FIM"]    = $intervalosMarcacoes[1]["IDDIAOCORRENCIA"];

                $horaEntrada    = $intervalosMarcacoes[0]["HORA_MINUTO"];
                $horaSaida      = $intervalosMarcacoes[1]["HORA_MINUTO"];
                $tempoIntervalo = DateUtil::subtrairHoras($horaSaida, $horaEntrada);
                $intervalos[$key]["TEMPO"] = $tempoIntervalo;

                $intervalos[$key]["IDREFERINTERVALO_INI"] = $intervalosMarcacoes[0]["IDREFERINTERVALO"];
                $intervalos[$key]["IDREFERINTERVALO_FIM"] = $intervalosMarcacoes[1]["IDREFERINTERVALO"];

                $identificaIntervalo = $this->verificaIdentificaIntervalo($intervalosMarcacoes[0]);
                if (isset($identificaIntervalo["NRPROGOCORRENCIA"])) {
                    $intervalos[$key]["ISALMOCO"] = "true";
                    $intervalos[$key]["NRPROGOCORRENCIA_ALMOCO"] = $identificaIntervalo["NRPROGOCORRENCIA"];
                } else {
                    $intervalos[$key]["ISALMOCO"] = "false";
                    $intervalos[$key]["NRPROGOCORRENCIA_ALMOCO"] = '';
                }

                $abonarSaida = $this->verificaAbonarSaida($intervalosMarcacoes[0]);

                if (isset($abonarSaida["NRPROGOCORRENCIA"])) {
                    $intervalos[$key]["ISJUSTIFICADA"] = "true";
                    $intervalos[$key]["NRPROGOCORRENCIA_JUSTIFICADA"]          = $abonarSaida["NRPROGOCORRENCIA"];
                    $intervalos[$key]["NRMOTIVOOCORFREQ_LABEL"]                = $abonarSaida["DSMOTIVOOCORFREQ"];
                    $intervalos[$key]["NRMOTIVOOCORFREQ"]                      = $abonarSaida["NRMOTIVOOCORFREQ"];
                    $intervalos[$key]["NRPROGOCORRENCIA_DSJUSTIFICATIVA"]      = $abonarSaida["DSOBSERVACAO"];
                    $intervalos[$key]["NRPROGOCORRENCIA_DSOBSERVACAORECUSA"]   = $abonarSaida["DSOBSERVACAORECUSA"];
                    $intervalos[$key]["NRPROGOCORRENCIA_IDPROGOCORCONFIRMADA"] = $abonarSaida["IDPROGOCORCONFIRMADA"];
                    $intervalos[$key]["NRPROGOCORRENCIA_DSPATH"]               = $abonarSaida["DSPATH"];

                    if (isset($abonarSaida["IDTIPOINTERVALO"]) && $abonarSaida["IDTIPOINTERVALO"] == self::IDTIPOINTERVALO_TRABALHO)
                        $intervalos[$key]["IDMARCACAO"] = self::LABEL_IDTIPOINTERVALO_TRABALHO;
                } else {
                    $intervalos[$key]["ISJUSTIFICADA"] = "false";
                    $intervalos[$key]["NRPROGOCORRENCIA_JUSTIFICADA"]          = '';
                    $intervalos[$key]["NRMOTIVOOCORFREQ_LABEL"]                = '';
                    $intervalos[$key]["NRMOTIVOOCORFREQ"]                      = '';
                    $intervalos[$key]["NRPROGOCORRENCIA_DSJUSTIFICATIVA"]      = '';
                    $intervalos[$key]["NRPROGOCORRENCIA_IDPROGOCORCONFIRMADA"] = '';
                    $intervalos[$key]["NRPROGOCORRENCIA_DSPATH"]               = '';
                }
            }
        }

        return $intervalos;

    }

    private function verificaIdentificaIntervalo($parameters) {
        $query = $this->entityManager->getConnection()->prepare("
            SELECT PROG.*
              FROM PTO_PROGOCORRENCIA PROG
              JOIN GPE_MARCACAOPONTO MARC
                ON MARC.NRMARCACAOPONTO = PROG.NRMARCACAOPONTO
               AND MARC.NRVINCULOM      = PROG.NRVINCULOM
               AND MARC.NRORG           = PROG.NRORG
             WHERE PROG.NRORG                 = :P_NRORG
               AND PROG.NRVINCULOM            = :P_NRVINCULOM
               AND PROG.NRTIPOPROGOCORRENCIA IN (20, 26)
               AND PROG.NRMARCACAOPONTO       = :P_NRMARCACAOPONTO
        ");

        $params = [
            'P_NRORG'           => $this->environment->getNrOrg(),
            'P_NRVINCULOM'      => $parameters['NRVINCULOM'],
            'P_NRMARCACAOPONTO' => $parameters['NRMARCACAOPONTO'],
        ];

        $query->execute($params);

        $ocorrencias = $query->fetchAll();

        if (isset($ocorrencias[0])) {
            return $ocorrencias[0];
        } else {
            return null;
        }
    }

    private function verificaAbonarSaida($parameters) {
        $query = $this->entityManager->getConnection()->prepare("
            SELECT
                *
              FROM PTO_PROGOCORRENCIA PROG
              LEFT JOIN PTO_MOTIVOOCORFREQ MOT
                ON MOT.NRMOTIVOOCORFREQ = PROG.NRMOTIVOOCORFREQ
                AND MOT.NRORG = PROG.NRORG
             WHERE PROG.NRORG                = :P_NRORG
               AND PROG.NRVINCULOM           = :P_NRVINCULOM
               AND PROG.NRTIPOPROGOCORRENCIA = 17
               AND PROG.NRMARCACAOPONTO      = :P_NRMARCACAOPONTO
        ");

        $query->execute(
            array(
                "P_NRORG"           => $this->environment->getNrOrg(),
                "P_NRVINCULOM"      => $parameters["NRVINCULOM"],
                "P_NRMARCACAOPONTO" => $parameters["NRMARCACAOPONTO"]
            )
        );

        $ocorrencias = $query->fetchAll();

        if (isset($ocorrencias[0])) {
            return $ocorrencias[0];
        } else {
            return null;
        }
    }

    public function deleteMarcacoesPonto($rows, $deleteAll = false, $dtapuracaoponto = null, $nrvinculom) {
        $nrorg      = $this->environment->getNrOrg();

        if ($deleteAll && !empty($dtapuracaoponto)) {

              $query = "DELETE FROM
                                GPE_MARCACAOPONTO
                              WHERE NRORG               = $nrorg
                                AND NRVINCULOM          = $nrvinculom
                                AND TRUNC(DTHRAJUSTADA) = TO_DATE('$dtapuracaoponto', 'DD/MM/YYYY')";

              $this->entityManager->getConnection()->exec($query);

        } else {
            foreach($rows as $row) {

                $nrmarcacaoponto_entrada = $row['NRMARCACAOPONTO'];
                $nrmarcacaoponto_saida   = $row['NRMARCACAOPONTO_SAIDA'];

                $query = "DELETE FROM
                                  PTO_PROGOCORRENCIA
                                WHERE NRORG             = $nrorg
                                  AND NRVINCULOM        = $nrvinculom
                                  AND NRMARCACAOPONTO  IN ($nrmarcacaoponto_entrada, $nrmarcacaoponto_saida)";

                $result_ocorrencia = $this->entityManager->getConnection()->exec($query);

                $query_marc = "DELETE FROM
                                       GPE_MARCACAOPONTO
                                     WHERE NRORG            = $nrorg
                                       AND NRVINCULOM       = $nrvinculom
                                       AND NRMARCACAOPONTO IN ($nrmarcacaoponto_entrada, $nrmarcacaoponto_saida)";

                $result_marcacao = $this->entityManager->getConnection()->exec($query_marc);
            }
        }

        //if ($result_marcacao > 0)
            return array('DELETE' => true);
        /*else
            return array('DELETE' => false);*/
    }
    
    /**
     * 
     * Rotina que valida o tempo trabalhado em um determinado dia, levantando exceções
     * 
     * @param string  $nrVinculo
     * @param string  $nrOrg
     * @param string  $dia
     * 
     * @throws Exception
     * 
     */
    public function validaTempoMarcacoes(
        string  $nrVinculo,
        string  $nrOrg,
        string  $dia
    ): bool {
        
        $params = [
            'P_NRORG'           => $nrOrg,
            'P_NRVINCULOM'      => $nrVinculo,
            'P_DTAPURACAOPONTO' => $dia
        ];

        // Busca os pares de marcação
        $paresMarcacoes = $this->getMarcacoesPonto($params);
        
        // Se não existirem pares, validar não é necessário
        if (empty($paresMarcacoes)) {
            return true;
        }
        
        // Busca os dados de horário do vínculo, para obter o IDTIPOHORARIO, necessário para a validação
        $dadosHorarioVinculo = $this->marcacaoRepository->getDadosHorarioVinculo($nrVinculo, $nrOrg, $dia);
        if (empty($dadosHorarioVinculo)) {
            throw new \Exception(Messages::VINC_HOR_DADOS_NOT_FOUND);
        }
        
        $idtipohorario = $dadosHorarioVinculo['IDTIPOHORARIO'];
        
        // Tipo do vínculo para validação
        $tipoVinculo = $this->vinculoService->getTipoVinculo($nrVinculo, $nrOrg);
        
        // Validação caso não seja estagiário
        if ($tipoVinculo != VinculoService::TIPO_VINC_EST) {
            
            $tempoTrabalhado    = 0;
            $tempoAlmoco        = 0;
            $almocoIdentificado = false;
            
            // Percorre os pares contabilizando o tempo as marcações
            foreach ($paresMarcacoes as $parMarcacao) {
                
                $identificacaoIntervalo = $parMarcacao['IDMARCACAO'];
                $intervaloAlmoco        = $parMarcacao['ISALMOCO'] == 'true';
                $ocorrenciaJustificada  = $parMarcacao['ISJUSTIFICADA'] == 'true';
                $nrmotivoocorfreq       = $parMarcacao['NRMOTIVOOCORFREQ'];
                $tempoIntervalo         = $this->calculaTempoIntervalo($parMarcacao['TEMPO']);
                
                if ($identificacaoIntervalo === self::LABEL_IDTIPOINTERVALO_TRABALHO) {
                    $tempoTrabalhado += $tempoIntervalo;
                } else if ($identificacaoIntervalo == self::LABEL_IDTIPOINTERVALO_AUSENCIA && $ocorrenciaJustificada && $nrmotivoocorfreq != OcorrenciaService::NRMOTIVOOCORFREQ_ATEST_COMP && !$intervaloAlmoco) {
                    $tempoTrabalhado += $tempoIntervalo;
                }
                
                if ($identificacaoIntervalo === self::LABEL_IDTIPOINTERVALO_AUSENCIA && $intervaloAlmoco) {
                    $almocoIdentificado = true;
                    $tempoAlmoco = $tempoIntervalo;
                }
                
            }
            
            // Trata os valores interpretados do dia
            if ($tempoTrabalhado > ApuracaoService::APUR_TEMPO_MIN_SEM_ALMOCO && $tempoTrabalhado <= ApuracaoService::APUR_TEMPO_MAXIMO_SEM_ALMOCO && !$almocoIdentificado) {
                throw new \Exception(Messages::APUR_HOR_TRAB_EXCEDIDO_1 . $dia);
            } else if (($tempoTrabalhado > ApuracaoService::APUR_TEMPO_MAXIMO_SEM_ALMOCO && !$almocoIdentificado) || ($tempoTrabalhado > ApuracaoService::APUR_TEMPO_MAXIMO_SEM_ALMOCO && $tempoAlmoco < ApuracaoService::APUR_TEMPO_MINIMO_ALMOCO && $idtipohorario == 'F')) {
                throw new \Exception(Messages::APUR_HOR_TRAB_EXCEDIDO_2 . $dia);
            }
        } else {

            // VALIDAR SE É FOLGA/FERIADO. CASO FOR, NÃO PRECISA VALIDAR
            if ($idtipohorario !== OcorrenciaService::IDTIPOHORARIO_FOLGA) {
                $dadosApuracao = $this->apuracaoRepository->getApuracaoPorData(
                    $nrVinculo,
                    $nrOrg,
                    $dia
                );
                
                if (empty($dadosApuracao)) {
                    throw new \Exception(Messages::APUR_DADOS_NOT_FOUND);
                }
                
                $saldoAnterior = isset($dadosApuracao['SALDO_BH_ANTERIOR']) || is_null($dadosApuracao['SALDO_BH_ANTERIOR']) ? $dadosApuracao['SALDO_BH_ANTERIOR'] : null;
                
                $tempoTrabalhado    = 0;
                $totalAusencias     = 0;
                $saldoAnterior      = floatval($saldoAnterior);
                $jornadaLimite      = ApuracaoService::APUR_JORNADA_EST + ($saldoAnterior * -1);
                
                foreach ($paresMarcacoes as $parMarcacao) {
                    
                    $ocorrenciaJustificada  = $parMarcacao['ISJUSTIFICADA'] == 'true';
                    $identificacaoIntervalo = $parMarcacao['IDMARCACAO'];
                    $tempoIntervalo         = $this->calculaTempoIntervalo($parMarcacao['TEMPO']);
                    
                    if ($identificacaoIntervalo === self::LABEL_IDTIPOINTERVALO_TRABALHO) {
                        $tempoTrabalhado += $tempoIntervalo;
                    } else if ($identificacaoIntervalo == self::LABEL_IDTIPOINTERVALO_AUSENCIA && $ocorrenciaJustificada) {
                        $tempoTrabalhado += $tempoIntervalo;
                    } else if ($identificacaoIntervalo == self::LABEL_IDTIPOINTERVALO_AUSENCIA && !$ocorrenciaJustificada) {
                        $totalAusencias += $tempoIntervalo;
                    }
                }
                
                // Estagiários tem um tempo de descanso definido em ApuracaoService::APUR_INTERVALO_JORNADA_EST
                if ($totalAusencias <= ApuracaoService::APUR_INTERVALO_JORNADA_EST) {
                    $tempoTrabalhado += $totalAusencias;
                } else {
                    $tempoTrabalhado += ApuracaoService::APUR_INTERVALO_JORNADA_EST;
                }
                
                if ($tempoTrabalhado > $jornadaLimite && !Utils::areFloatsEqual($tempoTrabalhado, $jornadaLimite)) {
                    
                    $diferencaString        = DateUtil::formataHoraDecimal($tempoTrabalhado - $jornadaLimite);
                    $jornadaLimiteFormatada = DateUtil::formataHoraDecimal($jornadaLimite);
                    $messageParams = array(
                        '@DIFERENCAJORNADA@'    => $diferencaString,
                        '@JORNADALIMITE@'       => $jornadaLimiteFormatada,
                        '@DIA@'                 => $dia,
                    );
                    throw new \Exception(strtr(Messages::APUR_HOR_TRAB_EXCEDIDO_EST, $messageParams));
                }
            }
        }

        return true;
    }
    
    /**
     *
     * Rotina que invoca a validação de tempo trabalhado para os dias de um determinado período. As validações se encontram em 'self::validaTempoMarcacoes'
     *
     * @param string  $nrVinculo
     * @param string  $nrOrg
     * @param string  $nrPeriodoApuracao
     *
     * @throws Exception
     */
    public function validaTempoMarcacoesPeriodo(
        string  $nrVinculo,
        string  $nrOrg,
        string  $nrPeriodoApuracao
    ) {
        $periodo = $this->periodoRepository->getPeriodo($nrOrg, $nrPeriodoApuracao);
        if (empty($periodo)) {
            throw new \Exception(Messages::PERIODO_PER_NOT_FOUND);
        }

        // Tranforma datas para datetimes
        $dataInicialPeriodo = DateUtil::getDataDeString($periodo['DTINICIALAPURACAO'], DateUtil::FORMATO_DATAHORA, true);
        $dataFinalPeriodo = DateUtil::getDataDeString($periodo['DTFINALAPURACAO'], DateUtil::FORMATO_DATAHORA, true);

        // Para cada dia do período, invoca a validação da rotina self::validaTempoMarcacoes
        $dataValidacao = $dataInicialPeriodo;
        while($dataValidacao <= $dataFinalPeriodo) {
            $this->validaTempoMarcacoes($nrVinculo, $nrOrg, $dataValidacao->format(DateUtil::FORMATO_DATA));
            $dataValidacao = DateUtil::adicionaIntervalo($dataValidacao, 1, DateUtil::DIA);
        }

        return true;
    }

    public function calculaTempoIntervalo(string $tempo) {
        $partesTempo = explode(':', $tempo);
        
        if (count($partesTempo) ==! 2) {
            return false;
        }
        
        $hora       = floatval($partesTempo[0]);
        $minutos    = floatval($partesTempo[1]);
        
        $tempoTotalEmHoras = $hora + ($minutos / 60);
        
        return $tempoTotalEmHoras;
        
    }
}