<?php

namespace Ems\Util;

class Time {
    public const GREATER_THAN   = 1;
    public const EQUAL_TO       = 0;
    public const LESS_THAN      = -1;

    /** @var int */
    private $minutes;

    public function __construct(
        int $minutes
    ) {
        $this->minutes = $minutes;
    }

    public function getMinutes(): int {
        return $this->minutes;
    }
    
    public function getHours(): float {
        return $this->getMinutes() / 60;
    }

    public function sum(
        Time $timeToSum
    ) {
        $sumInMinutes = $this->minutes + $timeToSum->getMinutes();
        $newTime = new Time($sumInMinutes);

        return $newTime;
    }

    public function sub(
        Time $timeToSub
    ) {
        $subInMinutes = $this->minutes - $timeToSub->getMinutes();
        $newTime = new Time($subInMinutes);

        return $newTime;
    }

    public function toString(): string {
        $minutesModulus = abs($this->minutes);

        $hours = floor($minutesModulus / 60);
        $minutes = $minutesModulus - $hours * 60;

        $hoursString = strval($hours);
        $minutesString = str_pad($minutes, 2, '0', STR_PAD_LEFT);

        $timeString = "$hoursString:$minutesString";

        if ($this->minutes < 0) {
            $timeString = "-$timeString";
        }

        return $timeString;
    }

    public function toFloat(): float {
        $timeFloat = floatval($this->minutes) / 60;

        return $timeFloat;
    }

    public function compare(
        Time $timeToCompare
    ): int {
        $minutesA = $this->minutes;
        $minutesB = $timeToCompare->getMinutes();

        if ($minutesA > $minutesB) {
            return Time::GREATER_THAN;
        }

        if ($minutesA < $minutesB) {
            return Time::LESS_THAN;
        }

        return Time::EQUAL_TO;
    }

    public function divide(
        float $divisor
    ): Time {
        $quocient = intval($this->minutes / $divisor);
        $time = new Time($quocient);

        return $time;
    }

    public function multiply(
        float $multiplier
    ): Time {
        $product = intval($this->minutes * $multiplier);
        $time = new Time($product);

        return $time;
    }

    public static function fromString(
        string $timeString
    ): Time {
        $isNegative = false;

        [$hours, $minutes] = explode(':', $timeString);

        if (strpos($hours, '-') !== false) {
            $isNegative = true;
            $hours = str_replace('-', '', $hours);
        }

        $timeInMinutes = intval($hours) * 60 + intval($minutes);

        if ($isNegative) {
            $timeInMinutes *= -1;
        }

        $newTime = new Time($timeInMinutes);

        return $newTime;
    }

    public static function fromFloat(
        float $timeDecimal
    ): Time {
        $minutes = round($timeDecimal * 60);
        $time = new Time($minutes);

        return $time;
    }
    
    public function diff(
        Time $timeToCompare
    ): Time {
        $minutesA = $this->minutes;
        $minutesB = $timeToCompare->getMinutes();
        
        $diffMinutes = $minutesA - $minutesB;
        
        return new Time($diffMinutes);
    }
}